# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Built-In Validation Plugins"""

# shortcuts:
from schemapack._internals.validation.plugins.content_schema import (
    ContentSchemaValidationPlugin,
)
from schemapack._internals.validation.plugins.expected_root import (
    ExpectedRootValidationPlugin,
)
from schemapack._internals.validation.plugins.missing_class import (
    MissingClassSlotValidationPlugin,
)
from schemapack._internals.validation.plugins.missing_origin import (
    MissingMandatoryOriginValidationPlugin,
)
from schemapack._internals.validation.plugins.missing_relations import (
    MissingRelationValidationPlugin,
)
from schemapack._internals.validation.plugins.missing_target import (
    MissingMandatoryTargetValidationPlugin,
)
from schemapack._internals.validation.plugins.multiple_target import (
    MultipleTargetValidationPlugin,
)
from schemapack._internals.validation.plugins.one_to_many_overlap import (
    TargetOverlapValidationPlugin,
)
from schemapack._internals.validation.plugins.unexpected_root import (
    UnexpectedRootValidationPlugin,
)
from schemapack._internals.validation.plugins.unknown_class import (
    UnknownClassSlotValidationPlugin,
)
from schemapack._internals.validation.plugins.unknown_relations import (
    UnknownRelationValidationPlugin,
)

__all__ = [
    "ContentSchemaValidationPlugin",
    "ExpectedRootValidationPlugin",
    "MissingClassSlotValidationPlugin",
    "MissingMandatoryOriginValidationPlugin",
    "MissingMandatoryTargetValidationPlugin",
    "MissingRelationValidationPlugin",
    "MultipleTargetValidationPlugin",
    "TargetIdValidationPlugin",
    "TargetOverlapValidationPlugin",
    "UnexpectedRootValidationPlugin",
    "UnknownClassSlotValidationPlugin",
    "UnknownRelationValidationPlugin",
]



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Tests for exporting schemapacks in mermaid format."""

import pytest

from schemapack import export_mermaid, load_schemapack
from tests.fixtures.examples import ERD_PATHS, VALID_SCHEMAPACK_PATHS


@pytest.mark.parametrize(
    "with_properties, file_suffix",
    [(True, "_w_props"), (False, "_wo_props")],
    ids=("content_props", "no_content_props"),
)
def test_export_mermaid(with_properties: bool, file_suffix: str):
    """Test dumping a schemapack in mermaid format."""
    example = "comprehensive_cardinalities_and_types"
    schemapack_path = VALID_SCHEMAPACK_PATHS[example]
    schemapack = load_schemapack(schemapack_path)

    erd_path = ERD_PATHS[example + file_suffix]

    observed_output = export_mermaid(
        schemapack=schemapack, content_properties=with_properties
    )
    assert observed_output == erd_path.read_text().strip()



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""A validation plugin."""

from schemapack._internals.validation.base import GlobalValidationPlugin
from schemapack.exceptions import ValidationPluginError
from schemapack.spec.datapack import DataPack
from schemapack.spec.schemapack import SchemaPack


class UnknownClassSlotValidationPlugin(GlobalValidationPlugin):
    """A global-scoped validation plugin validating that a datapack only contains
    slots for classes defined in the provided schemapack.
    """

    @staticmethod
    def does_apply(*, schemapack: SchemaPack) -> bool:
        """A classmethod to check whether this validation plugin is relevant for the
        given schemapack.

        Returns: True if this plugin is relevant for the given class definition.
        """
        return True

    def __init__(self, *, schemapack: SchemaPack):
        """This plugin is configured with the entire schemapack."""
        self._classes = set(schemapack.classes)

    def validate(self, *, datapack: DataPack):
        """Validate the entire datapack.

        Raises:
            schemapack.exceptions.ValidationPluginError: If validation fails.
        """
        unknown_classes = [
            class_ for class_ in datapack.resources if class_ not in self._classes
        ]

        if unknown_classes:
            raise ValidationPluginError(
                type_="UnknownClassSlotError",
                message=(
                    "Found slot(s) for class(es) not defined in the schemapack:"
                    + ", ".join(unknown_classes)
                ),
                details={
                    "unknown_classes": unknown_classes,
                },
            )



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""A validation plugin."""

from collections import Counter
from collections.abc import Mapping

from schemapack._internals.validation.base import ClassValidationPlugin
from schemapack.exceptions import ValidationPluginError
from schemapack.spec.custom_types import ResourceId
from schemapack.spec.datapack import DataPack, Resource
from schemapack.spec.schemapack import ClassDefinition


class TargetOverlapValidationPlugin(ClassValidationPlugin):
    """A class-scoped validation plugin validating that no overlap in one-to-*
    relations exists across resources of a schemapack class.
    This only applies to schemapack classes with one-to-* (multiple.origin is set to
    False) relations.
    """

    @staticmethod
    def does_apply(*, class_: ClassDefinition) -> bool:
        """A classmethod to check whether this validation plugin is relevant for the
        given class definition.

        Returns: True if this plugin is relevant for the given class definition.
        """
        return any(
            not relation.multiple.origin for relation in class_.relations.values()
        )

    def __init__(self, *, class_: ClassDefinition):
        """This plugin is configured with one specific class definition of a schemapack."""
        self._relations_of_interest = [
            name
            for name, relation in class_.relations.items()
            if not relation.multiple.origin
        ]

    def validate(
        self, *, class_resources: Mapping[ResourceId, Resource], datapack: DataPack
    ):
        """Validate all resources of a specific class. The entire datapack is provided
        for resolving relations to other classes.

        Raises:
            schemapack.exceptions.ValidationPluginError: If validation fails.
        """
        # Contains all overlapping target ids (values) per relation (keys) if any
        # overlaps are found for that relation:
        overlapping_ids_by_relation: dict[str, set[str]] = {}

        for relation_name in self._relations_of_interest:
            counter: Counter[str] = Counter()

            for resource in class_resources.values():
                try:
                    resource_target_ids = resource.relations[
                        relation_name
                    ].get_target_resources_as_set()
                except KeyError:
                    resource_target_ids = frozenset()

                counter.update(resource_target_ids)

            duplicate_target_ids = {k for k, v in counter.items() if v > 1}

            if duplicate_target_ids:
                overlapping_ids_by_relation[relation_name] = duplicate_target_ids

        if overlapping_ids_by_relation:
            raise ValidationPluginError(
                type_="CardinalityOverlapError",
                message=(
                    "Found overlapping target IDs for the following one-to-*"
                    + " relations:"
                    + ", ".join(overlapping_ids_by_relation)
                ),
                details={"overlapping_ids_by_relation": overlapping_ids_by_relation},
            )



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Logic to isolate a resource from a non-rooted datapack to create a rooted datapack.

Warning: This is an internal part of the library and might change without notice.
"""

from collections import defaultdict
from collections.abc import Mapping

from schemapack._internals.exceptions import (
    ClassNotFoundError,
    ResourceNotFoundError,
    SpecType,
)
from schemapack.exceptions import ValidationAssumptionError
from schemapack.spec.custom_types import ClassName, ResourceId
from schemapack.spec.datapack import DataPack, Resource
from schemapack.spec.schemapack import SchemaPack


def identify_resource_dependencies(
    *,
    datapack: DataPack,
    class_name: ClassName,
    resource_id: ResourceId,
    schemapack: SchemaPack,
    include_target: bool = False,
    _dependencies_by_class: dict[ClassName, set[ResourceId]] | None = None,
) -> dict[ClassName, set[ResourceId]]:
    """Identify all dependencies (recursively) for a given resource
    of the given class in the given datapack. Please note that it is assumed that
    the datapack has already been validated against the schemapack.

    Args:
        datapack:
            Containing the target resource and all dependencies.
        class_name:
            The class of the target resource.
        resource_id:
            The id of the target resource.
        schemapack:
            The schemapack used for looking up the classes of relations.
        include_target:
            Set to true if the target dependency shall be included in the result.
        _dependencies_by_class:
            A mapping containing resource ids (values) of dependencies by class names
            (keys) that are already in the recursion.

    Returns:
        A mapping containing resource ids (values) of dependencies by class names
        (keys).

    Raises:
        schemapack.Exceptions.ClassNotFoundError:
            If the class_name does not exist in the schemapack or datapack.
        schemapack.Exceptions.ResourceNotFoundError:
            If the resource_id does not exist in the schemapack or datapack.
        schemapack.Exceptions.ValidationAssumptionError:
            If it became apparent that the datapack was not already validated against
            the schemapack.
    """
    target_class_resources = datapack.resources.get(class_name)
    if target_class_resources is None:
        raise ClassNotFoundError(class_name=class_name, spec_type=SpecType.DATAPACK)

    target_resource = target_class_resources.get(resource_id)
    if target_resource is None:
        raise ResourceNotFoundError(class_name=class_name, resource_id=resource_id)

    target_class_definition = schemapack.classes.get(class_name)
    if target_class_definition is None:
        raise ClassNotFoundError(class_name=class_name, spec_type=SpecType.SCHEMAPACK)

    if _dependencies_by_class is None:
        _dependencies_by_class = defaultdict(set)
    else:  # do not remove the target in recursion
        include_target = True
    _dependencies_by_class[class_name].add(resource_id)

    for relation_name in target_resource.relations:
        try:
            target_class_name = target_class_definition.relations[
                relation_name
            ].targetClass
        except KeyError as error:
            raise ValidationAssumptionError(
                context="relation resolution in schemapack"
            ) from error

        try:
            target_ids = target_resource.relations[
                relation_name
            ].get_target_resources_as_set()
        except KeyError as error:
            raise ValidationAssumptionError(
                context="relation resolution in datapack"
            ) from error

        # Remove already known dependencies to avoid circular loops
        target_ids -= _dependencies_by_class.get(target_class_name, set())

        for target_id in target_ids:
            _dependencies_by_class[target_class_name].add(target_id)

            # Recursively add dependencies of this target resource:
            identify_resource_dependencies(
                datapack=datapack,
                class_name=target_class_name,
                resource_id=target_id,
                schemapack=schemapack,
                _dependencies_by_class=_dependencies_by_class,
            )

    if not include_target:  # only removed in the initial call
        _dependencies_by_class[class_name].discard(resource_id)

    return _dependencies_by_class


def downscope_datapack(
    datapack: DataPack,
    resource_map: Mapping[ClassName, set[ResourceId]],
    ignore_non_existing: bool = False,
) -> DataPack:
    """Downscope a datapack to only contain the given resources.

    Args:
        ignore_non_existing:
            Controls how to handle resources from the resource_map that are not in the
            datapack. If set to `True`, these resources will be ignored. If set to
            `False` (default), a KeyError will be raised.

    Raises:
        KeyError:
            If a resource from the resource_map is not in the datapack and
            ignore_non_existing is set to `False`.
    """
    resources: dict[ClassName, dict[ResourceId, Resource]] = defaultdict(dict)

    for class_name, resource_ids in resource_map.items():
        try:
            existing_resources = datapack.resources[class_name]
        except KeyError as error:
            if ignore_non_existing:
                continue
            raise error

        for resource_id in resource_ids:
            try:
                existing_resource = existing_resources[resource_id]
            except KeyError as error:
                if ignore_non_existing:
                    continue
                raise error

            resources[class_name][resource_id] = existing_resource

    return datapack.model_copy(update={"resources": resources})


def isolate_resource(
    *,
    datapack: DataPack,
    class_name: ClassName,
    resource_id: ResourceId,
    schemapack: SchemaPack,
) -> DataPack:
    """Isolate a resource from a non-rooted datapack to created a rooted datapack. I.e.
    the resulting datapack will only contain resources referenced by the root resource
    as well as the root resource itself.

    Please note:
        The returned rooted datapack will not be compatible anymore with the original
        non-rooted schemapack.

    Raises:
        schemapack.Exceptions.ClassNotFoundError:
            If the class_name does not exist in the schemapack or datapack.
        schemapack.Exceptions.ResourceNotFoundError:
            If the resource_id does not exist in the schemapack or datapack.
        schemapack.Exceptions.ValidationAssumptionError:
            If it became apparent that the datapack was not already validated against
            the schemapack.
    """
    dependency_map = identify_resource_dependencies(
        datapack=datapack,
        class_name=class_name,
        resource_id=resource_id,
        schemapack=schemapack,
        include_target=True,
    )
    rooted_datapack = downscope_datapack(datapack=datapack, resource_map=dependency_map)
    rooted_datapack = rooted_datapack.model_copy(
        update={"rootResource": resource_id, "rootClass": class_name}
    )
    return rooted_datapack


def identify_class_dependencies(
    *,
    class_name: ClassName,
    schemapack: SchemaPack,
    _class_blacklist: set[ClassName] | None = None,
) -> set[ClassName]:
    """Identify all dependencies (recursively) for a given class in the given schemapack.

    Args:
        class_name:
            The class for which to identify dependencies.
        schemapack:
            The schemapack used for looking up the classes of relations.
        _class_blacklist:
            A set of class names to avoid getting lost in infinity loop for circular
            dependencies. This is only used internally for recursion.

    Raises:
        schemapack.Exceptions.ClassNotFoundError:
            If the class_name does not exist in the schemapack.
    """
    class_definition = schemapack.classes.get(class_name)
    if class_definition is None:
        raise ClassNotFoundError(class_name=class_name, spec_type=SpecType.SCHEMAPACK)

    dependencies: set[ClassName] = set()

    for relation in class_definition.relations.values():
        if _class_blacklist and relation.targetClass in _class_blacklist:
            continue

        dependencies.add(relation.targetClass)

        nested_dependencies = identify_class_dependencies(
            class_name=relation.targetClass,
            schemapack=schemapack,
            _class_blacklist=dependencies,
        )
        dependencies.update(nested_dependencies)

    return dependencies


def downscope_schemapack(
    *, schemapack: SchemaPack, classes_to_keep: set[ClassName]
) -> SchemaPack:
    """Downscope a schemapack to only contain the given classes.

    Raises:
        schemapack.Exceptions.ClassNotFoundError:
            If one of the classes in classes_to_keep does not exist in the schemapack.
    """
    try:
        downscoped_classes = {
            class_name: schemapack.classes[class_name] for class_name in classes_to_keep
        }
    except KeyError as error:
        raise ClassNotFoundError(
            class_name=error.args[0], spec_type=SpecType.SCHEMAPACK
        ) from error

    return schemapack.model_copy(update={"classes": downscoped_classes})


def isolate_class(*, class_name: ClassName, schemapack: SchemaPack) -> SchemaPack:
    """Return a copy of the provided schemapack that is rooted to the specified class.


    Raises:
        schemapack.Exceptions.ClassNotFoundError:
            If the class_name does not exist in the schemapack or datapack.
    """
    dependencies = identify_class_dependencies(
        class_name=class_name, schemapack=schemapack
    )
    dependencies.add(class_name)
    schemapack = downscope_schemapack(
        schemapack=schemapack, classes_to_keep=dependencies
    )

    return schemapack.model_copy(update={"rootClass": class_name})


def isolate(
    *,
    root_class_name: ClassName,
    root_resource_id: ResourceId,
    schemapack: SchemaPack,
    datapack: DataPack,
) -> tuple[SchemaPack, DataPack]:
    """Create copies of the provided schemapacks and datapacks that are rooted towards
    the specified class and resource. I.e. the resulting datapack will only contain
    resources referenced by the root resource as well as the root resource itself.

    Returns:
        A tuple containing both the rooted schemapack and the rooted datapack.

    Raises:
        schemapack.Exceptions.ClassNotFoundError:
            If the class_name does not exist in the schemapack or datapack.
        schemapack.Exceptions.ResourceNotFoundError:
            If the resource_id does not exist in the schemapack or datapack.
        schemapack.Exceptions.ValidationAssumptionError:
            If it became apparent that the datapack was not already validated against
            the schemapack.
    """
    rooted_schemapack = isolate_class(class_name=root_class_name, schemapack=schemapack)
    rooted_datapack = isolate_resource(
        datapack=datapack,
        class_name=root_class_name,
        resource_id=root_resource_id,
        schemapack=schemapack,
    )
    return rooted_schemapack, rooted_datapack



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Tests the models module."""

import json

from immutabledict import immutabledict

from schemapack import load_schemapack
from schemapack._internals.load import load_datapack
from schemapack._internals.utils import read_json_or_yaml_mapping
from schemapack.spec.datapack import (
    SUPPORTED_DATA_PACK_VERSIONS,
    DataPack,
)
from tests.fixtures.examples import VALID_DATAPACK_PATHS, VALID_SCHEMAPACK_PATHS


def test_schemapack_is_hashable():
    """Test that instances of SchemaPack are hashable."""
    schemapack = load_schemapack(VALID_SCHEMAPACK_PATHS["simple_relations"])
    _ = hash(schemapack)


def test_comparison_and_hashing():
    """Test that equivalent schemapack (content schemas embedded or not) have the same
    hash and are equal.
    """
    schemapack = load_schemapack(VALID_SCHEMAPACK_PATHS["simple_relations"])
    schemapack_condensed = load_schemapack(
        VALID_SCHEMAPACK_PATHS["simple_relations_condensed"]
    )

    assert hash(schemapack) == hash(schemapack_condensed)
    assert schemapack == schemapack_condensed


def test_comparison_and_hashing_different():
    """Test that different SchemaPack instances have different hashes and are unequal."""
    schemapack = load_schemapack(VALID_SCHEMAPACK_PATHS["simple_relations"])

    schemapack_modified = schemapack.model_copy(
        update={
            "classes": immutabledict(
                {
                    "AdditionalClass": schemapack.classes[
                        set(schemapack.classes).pop()
                    ],
                    **schemapack.classes,
                }
            )
        }
    )

    assert hash(schemapack) != hash(schemapack_modified)
    assert schemapack != schemapack_modified


def test_content_schema_serialization():
    """Test that content schemas of a schemapack are serialized as dicts."""
    schemapack_path = VALID_SCHEMAPACK_PATHS["simple_relations_condensed"]
    schemapack = load_schemapack(schemapack_path)
    expected_schemapack_dict = read_json_or_yaml_mapping(schemapack_path)

    serialized_schemapack = json.loads(schemapack.model_dump_json())

    for class_name in schemapack.classes:
        assert (
            serialized_schemapack["classes"][class_name]["content"]
            == expected_schemapack_dict["classes"][class_name]["content"]
        )


def test_datapack_is_hashable():
    """Test that instances of DataPack are hashable."""
    datapack = load_datapack(VALID_DATAPACK_PATHS["simple_relations.simple_resources"])
    _ = hash(datapack)


def test_datapack_target_id_ordering_upon_dump():
    """Test that target_ids of a resource relation are sorted (i.e. the output is
    predictable) when serializing a datapack.
    """
    unsorted_target_ids = ["c", "a", "b"]
    sorted_target_ids = sorted(unsorted_target_ids)

    datapack = DataPack.model_validate(
        {
            "datapack": SUPPORTED_DATA_PACK_VERSIONS[-1],
            "resources": {
                "TestClass": {
                    "test_resource": {
                        "content": {},
                        "relations": {
                            "test_relations": {
                                "targetClass": "SomeClass",
                                "targetResources": unsorted_target_ids,
                            }
                        },
                    }
                },
                "SomeClass": {
                    "a": {"content": {}, "relations": {}},
                    "b": {"content": {}, "relations": {}},
                    "c": {"content": {}, "relations": {}},
                },
            },
        }
    )

    serialize_datapack = json.loads(datapack.model_dump_json())

    assert (
        serialize_datapack["resources"]["TestClass"]["test_resource"]["relations"][
            "test_relations"
        ]["targetResources"]
        == sorted_target_ids
    )



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Test the isolate module."""

import pytest

from schemapack import (
    exceptions,
    isolate,
    isolate_class,
    isolate_resource,
    load_datapack,
    load_schemapack,
)
from schemapack.spec.custom_types import ClassName, ResourceId
from tests.fixtures.examples import VALID_DATAPACK_PATHS, VALID_SCHEMAPACK_PATHS


@pytest.mark.parametrize(
    "datapack_name, resource_class, resource_id, rooted_datapack_name",
    [
        (
            "self_relation.multiple_relation_groups",
            "SomeClass",
            "a",
            "self_relation_rooted.rooted_nested_relations",
        ),
        (
            "simple_relations.simple_resources",
            "Dataset",
            "example_dataset_1",
            "simple_relations_rooted.rooted_simple_resources",
        ),
        (
            "self_relation.circular_relations",
            "SomeClass",
            "a",
            "self_relation_rooted.rooted_circular_relations",
        ),
        (
            "self_relation.circular_self_relations",
            "SomeClass",
            "a",
            "self_relation_rooted.rooted_circular_self_relations",
        ),
        (
            "complex_cardinality.all_cardinalities",
            "A",
            "a1",
            "complex_cardinality_rooted.rooted_all_cardinalities",
        ),
    ],
    ids=[
        "nested_relations",
        "cross_class_relations",
        "circular_relations",
        "circular_self_relations",
        "all_cardinalities",
    ],
)
def test_isolate(
    datapack_name: str,
    resource_class: ClassName,
    resource_id: ResourceId,
    rooted_datapack_name: str,
):
    """Test the isolate function."""
    datapack_path = VALID_DATAPACK_PATHS[datapack_name]
    expected_rooted_datapack_path = VALID_DATAPACK_PATHS[rooted_datapack_name]
    schemapack_name = datapack_name.split(".", 1)[0]
    schemapack_path = VALID_SCHEMAPACK_PATHS[schemapack_name]
    rooted_schemapack_name = rooted_datapack_name.split(".", 1)[0]
    expected_rooted_schemapack_path = VALID_SCHEMAPACK_PATHS[rooted_schemapack_name]

    datapack = load_datapack(datapack_path)
    expected_rooted_datapack = load_datapack(expected_rooted_datapack_path)
    schemapack = load_schemapack(schemapack_path)
    expected_rooted_schemapack = load_schemapack(expected_rooted_schemapack_path)

    rooted_schemapack, rooted_datapack = isolate(
        root_class_name=resource_class,
        root_resource_id=resource_id,
        schemapack=schemapack,
        datapack=datapack,
    )

    assert rooted_schemapack == expected_rooted_schemapack
    assert rooted_datapack == expected_rooted_datapack


def test_isolate_with_non_existing_class():
    """Test the isolate function with a non-existing class."""
    schemapack_path = VALID_SCHEMAPACK_PATHS["simple_relations"]
    datapack_path = VALID_DATAPACK_PATHS["simple_relations.simple_resources"]
    schemapack = load_schemapack(schemapack_path)
    datapack = load_datapack(datapack_path)

    with pytest.raises(exceptions.ClassNotFoundError) as error:
        isolate(
            root_class_name="NonExistingClass",
            root_resource_id="example_dataset_1",
            schemapack=schemapack,
            datapack=datapack,
        )
    assert "schemapack" in str(error)


def test_isolate_with_non_existing_resource():
    """Test the isolate function with a non-existing resource."""
    schemapack_path = VALID_SCHEMAPACK_PATHS["simple_relations"]
    datapack_path = VALID_DATAPACK_PATHS["simple_relations.simple_resources"]
    schemapack = load_schemapack(schemapack_path)
    datapack = load_datapack(datapack_path)

    with pytest.raises(exceptions.ResourceNotFoundError):
        isolate(
            root_class_name="Dataset",
            root_resource_id="NonExistingResource",
            schemapack=schemapack,
            datapack=datapack,
        )


def test_isolate_resource_non_existing_class():
    """Test the isolate_resource function with a non-existing class. Happy paths are
    tested as part of the isolate function and appear not worth repeating for this
    specific function.
    """
    schemapack_path = VALID_SCHEMAPACK_PATHS["simple_relations"]
    datapack_path = VALID_DATAPACK_PATHS["simple_relations.simple_resources"]
    schemapack = load_schemapack(schemapack_path)
    datapack = load_datapack(datapack_path)

    with pytest.raises(exceptions.ClassNotFoundError):
        isolate_resource(
            class_name="NonExistingClass",
            resource_id="example_dataset_1",
            schemapack=schemapack,
            datapack=datapack,
        )


def test_isolate_class_non_existing_class():
    """Test the isolate_class function with a non-existing class. Happy paths are
    tested as part of the isolate function and appear not worth repeating for this
    specific function.
    """
    schemapack_path = VALID_SCHEMAPACK_PATHS["simple_relations"]
    schemapack = load_schemapack(schemapack_path)

    with pytest.raises(exceptions.ClassNotFoundError):
        isolate_class(
            schemapack=schemapack,
            class_name="NonExistingClass",
        )


def test_isolate_class_downscoping():
    """Test that unrelated classes are not included in the isolated schemapack."""
    schemapack = load_schemapack(VALID_SCHEMAPACK_PATHS["unrelated_classes"])
    expected_schemapack = load_schemapack(
        VALID_SCHEMAPACK_PATHS["unrelated_classes_rooted"]
    )

    observed_schemapack = isolate_class(class_name="SomeClass", schemapack=schemapack)
    assert observed_schemapack == expected_schemapack



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Handling user-derived exceptions in the CLI."""

from contextlib import contextmanager

import typer

from schemapack import exceptions
from schemapack._internals.cli import exit_codes
from schemapack._internals.cli.printing import print_exception, print_final_failure


@contextmanager
def expect_validation_errors():
    """Handle validation errors."""
    try:
        yield
    except exceptions.ValidationError as error:
        print_exception(error, exception_name="ValidationError")
        print_final_failure(
            "The provided datapack is not valid wrt the provided schemapack.",
        )
        raise typer.Exit(exit_codes.VALIDATION_ERROR) from None


@contextmanager
def expect_datapackspec_errors():
    """Handle datapackspec errors."""
    try:
        yield
    except exceptions.DataPackSpecError as error:
        print_exception(error, exception_name="DataPackSpecError")
        print_final_failure(
            "The provided document did not comply with the specs of a datapack."
        )
        raise typer.Exit(exit_codes.DATAPACK_SPEC_ERROR) from None


@contextmanager
def expect_schemapack_errors():
    """Handle schemapack errors."""
    try:
        yield
    except exceptions.SchemaPackSpecError as error:
        print_exception(error, exception_name="SchemaPackSpecError")
        print_final_failure(
            "The provided document did not comply with the specs of a schemapack."
        )
        raise typer.Exit(exit_codes.SCHEMAPACK_SPEC_ERROR) from None


@contextmanager
def expect_common_user_errors():
    """Handle all user-derived errors."""
    with (
        expect_validation_errors(),
        expect_datapackspec_errors(),
        expect_schemapack_errors(),
    ):
        yield


@contextmanager
def expect_class_not_found_errors():
    """Handle class not found errors."""
    try:
        yield
    except exceptions.ClassNotFoundError as error:
        print_exception(error, exception_name="ClassNotFoundError")
        print_final_failure(
            "The provided document did not contain the specified class."
        )
        raise typer.Exit(exit_codes.CLASS_NOT_FOUND_ERROR) from None


@contextmanager
def expect_resource_not_found_errors():
    """Handle resource not found errors."""
    try:
        yield
    except exceptions.ResourceNotFoundError as error:
        print_exception(error, exception_name="ResourceNotFoundError")
        print_final_failure(
            "The provided document did not contain the specified resource."
        )
        raise typer.Exit(exit_codes.RESOURCE_NOT_FOUND_ERROR) from None


@contextmanager
def expect_class_or_resource_not_found_errors():
    """Handle class or resource not found errors."""
    with expect_class_not_found_errors(), expect_resource_not_found_errors():
        yield



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test schema comparison functionality."""

from dataclasses import dataclass
from itertools import combinations
from pathlib import Path

import pytest
from arcticfreeze import FrozenDict

from schemapack import (
    is_equal_schemapack,
    is_equivalent_schemapack,
    load_schemapack,
)
from schemapack._internals.compare import (
    compare_class_relations_semantically,
    compare_relations_semantically,
)
from schemapack._internals.exceptions import (
    InequivalentContentSchemas,
    InequivalentSchemapacks,
)
from schemapack._internals.spec.schemapack import (
    ClassRelation,
    MandatoryRelationSpec,
    MultipleRelationSpec,
)
from tests.fixtures.examples import (
    COMPARISON_SCHEMAPACK_PATHS,
    REPRESENTATIVE_SCHEMAPACK_PATHS,
    SCHEMAPACK_PAIRED_COMPARISON_PATHS,
)

# Expected comparison results for each test case
# True indicates schemapacks should be considered equal
EXPECTED_RESULTS = {
    "equal_schemas_optional_untyped_property": True,
    "equal_schemas_with_relations": True,
    "unequal_schemas_different_classes": False,
    "unequal_schemas_different_relation_spec": False,
    "unequal_schemas_different_content": False,
}

EXPECTED_EXCEPTIONS = {
    "equal_schemas_optional_untyped_property": None,
    "equal_schemas_with_relations": None,
    "unequal_schemas_different_classes": InequivalentSchemapacks,
    "unequal_schemas_different_relation_spec": InequivalentContentSchemas,
    "unequal_schemas_different_content": InequivalentContentSchemas,
}


@dataclass
class _Relation:
    """Defines constraints and cardinality rules for a test relation."""

    target_class: str
    mandatory_origin: bool = True
    mandatory_target: bool = True
    multiple_origin: bool = True
    multiple_target: bool = True

    def _make_class_relation(self) -> ClassRelation:
        return ClassRelation(
            targetClass=self.target_class,
            mandatory=MandatoryRelationSpec(
                origin=self.mandatory_origin, target=self.mandatory_target
            ),
            multiple=MultipleRelationSpec(
                origin=self.multiple_origin, target=self.multiple_target
            ),
        )


def _relation(*relations: tuple[str, _Relation]) -> FrozenDict:
    """Create one or more ClassRelation entries."""
    return FrozenDict({name: rel._make_class_relation() for name, rel in relations})


@pytest.mark.parametrize(
    "relations1, relations2",
    [
        pytest.param(
            FrozenDict(),
            _relation(("a", _Relation(target_class="B"))),
            id="one_empty",
        ),
        pytest.param(
            _relation(("a", _Relation(target_class="B"))),
            _relation(("b", _Relation(target_class="B"))),
            id="different_keys",
        ),
        pytest.param(
            _relation(("a", _Relation(target_class="B"))),
            _relation(("a", _Relation(target_class="C"))),
            id="different_target_class",
        ),
        pytest.param(
            _relation(
                ("a", _Relation(target_class="B", mandatory_origin=True)),
                ("c", _Relation(target_class="C")),
            ),
            _relation(("a", _Relation(target_class="B", mandatory_origin=True))),
            id="different_relation_counts",
        ),
    ],
)
def test_compare_relations_unhappy(relations1: FrozenDict, relations2: FrozenDict):
    """Test semantic equality of relation mappings for various cases."""
    assert not compare_relations_semantically(relations1, relations2)


def test_compare_empty_relations():
    """Test that empty relations are considered semantically equal."""
    assert compare_relations_semantically(FrozenDict(), FrozenDict())


def test_compare_class_relations_happy():
    """Test that identical ClassRelation objects are equal."""
    relation1 = _Relation(target_class="B")._make_class_relation()
    relation2 = _Relation(target_class="B")._make_class_relation()
    assert compare_class_relations_semantically(relation1, relation2)


@pytest.mark.parametrize(
    "relation1, relation2",
    [
        pytest.param(
            _Relation(target_class="B")._make_class_relation(),
            _Relation(target_class="C")._make_class_relation(),
            id="different_target_class",
        ),
        pytest.param(
            _Relation(target_class="B", mandatory_origin=True)._make_class_relation(),
            _Relation(target_class="B", mandatory_origin=False)._make_class_relation(),
            id="different_mandatory_origin",
        ),
        pytest.param(
            _Relation(target_class="B", mandatory_target=True)._make_class_relation(),
            _Relation(target_class="B", mandatory_target=False)._make_class_relation(),
            id="different_mandatory_target",
        ),
        pytest.param(
            _Relation(target_class="B", multiple_origin=True)._make_class_relation(),
            _Relation(target_class="B", multiple_origin=False)._make_class_relation(),
            id="different_multiple_origin",
        ),
        pytest.param(
            _Relation(target_class="B", multiple_target=True)._make_class_relation(),
            _Relation(target_class="B", multiple_target=False)._make_class_relation(),
            id="different_multiple_target",
        ),
    ],
)
def test_compare_class_relations_unhappy(
    relation1: ClassRelation, relation2: ClassRelation
):
    """Test semantic equality of ClassRelation objects."""
    assert not compare_class_relations_semantically(relation1, relation2)


@pytest.mark.parametrize(
    "path", COMPARISON_SCHEMAPACK_PATHS.values(), ids=COMPARISON_SCHEMAPACK_PATHS
)
def test_equivalence_reflexivity(path: Path):
    """
    Verify reflexivity of schemapack equivalence.

    A schemapack must be equivalent to itself.
    """
    schema = load_schemapack(path)
    assert is_equivalent_schemapack(schema, schema)


@pytest.mark.parametrize(
    "name, path1, path2",
    SCHEMAPACK_PAIRED_COMPARISON_PATHS,
    ids=[
        f"{name}: {path1.name} vs {path2.name}"
        for name, path1, path2 in SCHEMAPACK_PAIRED_COMPARISON_PATHS
    ],
)
def test_equivalence_symmetry_and_transitivity(name: str, path1: Path, path2: Path):
    """
    Verify symmetry of schemapack equivalence on known equivalent pairs.

    For each known equivalent pair (A, B), equivalence must hold in both
    directions: A ≡ B and B ≡ A.

    Transitivity is covered implicitly by the completeness of the equivalent
    pair set used for parametrization.
    """
    schema1 = load_schemapack(path1)
    schema2 = load_schemapack(path2)
    assert is_equivalent_schemapack(schema1, schema2)
    assert is_equivalent_schemapack(schema2, schema1)


@pytest.mark.parametrize(
    "path1, path2",
    list(combinations(REPRESENTATIVE_SCHEMAPACK_PATHS.values(), 2)),
    ids=[
        f"{path1.parent.name}.{path1.stem.split('.')[0]} vs {path2.parent.name}.{path2.stem.split('.')[0]}"
        for path1, path2 in combinations(REPRESENTATIVE_SCHEMAPACK_PATHS.values(), 2)
    ],
)
def test_inequivalence(path1: Path, path2: Path):
    """
    Verify inequivalence across representative schemapacks from different test cases.

    Any two schemapacks chosen from different representative scenarios must
    not be considered equivalent.
    """
    schema1 = load_schemapack(path1)
    schema2 = load_schemapack(path2)
    assert not is_equivalent_schemapack(schema1, schema2)


def test_equal_schemapack_with_identical_file_loaded_twice():
    """Test that the same file loaded twice produces equal objects."""
    path = COMPARISON_SCHEMAPACK_PATHS["ghga_ingress.ingress"]

    schema1 = load_schemapack(path)
    schema2 = load_schemapack(path)

    assert is_equal_schemapack(schema1, schema2)


def test_structurally_different_schemapacks():
    """Testcase where semantically equivalent schemapacks are not structurally equal."""
    path1 = COMPARISON_SCHEMAPACK_PATHS["all_mandatory.some_descriptions"]
    path2 = COMPARISON_SCHEMAPACK_PATHS["all_mandatory.no_descriptions"]

    schema1 = load_schemapack(path1)
    schema2 = load_schemapack(path2)

    assert not is_equal_schemapack(schema1, schema2)



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Describes a command line interface"""

from pathlib import Path
from typing import Annotated

import typer

from schemapack import __version__
from schemapack._internals.cli import exit_codes
from schemapack._internals.cli.exception_handling import (
    expect_class_or_resource_not_found_errors,
    expect_common_user_errors,
    expect_datapackspec_errors,
    expect_schemapack_errors,
)
from schemapack._internals.cli.printing import (
    print_final_success,
    print_output,
)
from schemapack._internals.dump import dumps_datapack, dumps_schemapack
from schemapack._internals.erd import export_mermaid as export_mermaid_impl
from schemapack._internals.isolate import isolate_class as isolate_class_impl
from schemapack._internals.isolate import isolate_resource as isolate_resource_impl
from schemapack._internals.load import load_datapack, load_schemapack
from schemapack._internals.main import load_and_validate

cli = typer.Typer()


def version_callback(
    version: bool = False,
):
    if version:
        print_output(__version__)
        raise typer.Exit(exit_codes.SUCCESS)


@cli.callback()
def common(
    ctx: typer.Context,
    version: Annotated[
        bool,
        typer.Option(
            "--version",
            callback=version_callback,
            help="Show the version of the library and exit.",
        ),
    ] = False,
):
    """Common arguments and options."""
    pass


@cli.command()
def validate(
    *,
    schemapack: Annotated[
        Path,
        typer.Option(
            "--schemapack",
            "-s",
            help="Provide the path to a schemapack to validate against.",
        ),
    ],
    datapack: Annotated[
        Path,
        typer.Option(
            "--datapack", "-d", help="Provide the path to a datapack to validate."
        ),
    ],
):
    """Validate a datapack against a schemapack."""
    with expect_common_user_errors():
        load_and_validate(schemapack_path=schemapack, datapack_path=datapack)

    print_final_success("The provided datapack is valid wrt the provided schemapack.")


@cli.command()
def check_schemapack(
    *,
    schemapack: Annotated[
        Path,
        typer.Argument(
            help="Provide the path to a JSON/YAML file to check against the specs.",
        ),
    ],
):
    """Check if the provided JSON/YAML document complies with the schemapack specs."""
    with expect_schemapack_errors():
        load_schemapack(schemapack)

    print_final_success(
        "The provided document complies with the specs of a schemapack."
    )


@cli.command()
def check_datapack(
    *,
    datapack: Annotated[
        Path,
        typer.Argument(
            help="Provide the path to a JSON/YAML file to check against the specs.",
        ),
    ],
):
    """Check if the provided JSON/YAML document complies with the datapack specs."""
    with expect_datapackspec_errors():
        load_datapack(datapack)

    print_final_success("The provided document complies with the specs of a datapack.")


@cli.command()
def condense_schemapack(
    *,
    schemapack: Annotated[
        Path,
        typer.Argument(
            help="Provide the path to a JSON/YAML file to check against the specs.",
        ),
    ],
    json: Annotated[
        bool,
        typer.Option(
            "--json",
            "-j",
            help="Output the condensed schemapack as JSON instead of YAML.",
        ),
    ] = False,
):
    """Writes a version of the provided schemapack with embedded content schemas to
    stdout.
    """
    with expect_schemapack_errors():
        schemapack_dict = load_schemapack(schemapack)
        condensed_schemapack = dumps_schemapack(schemapack_dict, yaml_format=not json)
        print_output(condensed_schemapack)


@cli.command()
def isolate_resource(
    *,
    schemapack: Annotated[
        Path,
        typer.Option(
            "--schemapack",
            "-s",
            help=(
                "The path to a schemapack that describes the structure of the"
                " provided input datapack (not the rooted datapack output by this"
                " command)."
            ),
        ),
    ],
    datapack: Annotated[
        Path,
        typer.Option(
            "--datapack",
            "-d",
            help="The path to a datapack from which the resource shall be isolated.",
        ),
    ],
    class_name: Annotated[
        str,
        typer.Option(
            "--class-name",
            "-c",
            help="The name of the class of the resource to isolate.",
        ),
    ],
    resource_id: Annotated[
        str,
        typer.Option(
            "--resource-id",
            "-r",
            help="The ID of the resource to isolate.",
        ),
    ],
    json: Annotated[
        bool,
        typer.Option(
            "--json",
            "-j",
            help="Output the rooted datapack as JSON instead of YAML.",
        ),
    ] = False,
):
    """Isolate a resource from the given datapack and write a datapack that is rooted to
    this resource to stdout.
    """
    with expect_common_user_errors():
        schemapack_obj, datapack_obj = load_and_validate(
            schemapack_path=schemapack, datapack_path=datapack
        )

    with expect_class_or_resource_not_found_errors():
        rooted_datapack = isolate_resource_impl(
            datapack=datapack_obj,
            class_name=class_name,
            resource_id=resource_id,
            schemapack=schemapack_obj,
        )

    rooted_datapack_str = dumps_datapack(rooted_datapack, yaml_format=not json)
    print_output(rooted_datapack_str)


@cli.command()
def isolate_class(
    *,
    schemapack: Annotated[
        Path,
        typer.Option(
            "--schemapack",
            "-s",
            help=(
                "The path to a schemapack that describes the structure of the"
                " provided input datapack (not the rooted datapack output by this"
                " command)."
            ),
        ),
    ],
    class_name: Annotated[
        str,
        typer.Option(
            "--class-name",
            "-c",
            help="The name of the class of the resource to isolate.",
        ),
    ],
    json: Annotated[
        bool,
        typer.Option(
            "--json",
            "-j",
            help="Output the rooted datapack as JSON instead of YAML.",
        ),
    ] = False,
):
    """Isolate a class from the given schemapack and write a condensed (with content
    schemas being embedded) schemapack that is rooted to this class to stdout.
    """
    with expect_common_user_errors():
        schemapack_obj = load_schemapack(schemapack)

    with expect_class_or_resource_not_found_errors():
        rooted_schemapack = isolate_class_impl(
            schemapack=schemapack_obj,
            class_name=class_name,
        )

    rooted_schemapack_str = dumps_schemapack(rooted_schemapack, yaml_format=not json)
    print_output(rooted_schemapack_str)


@cli.command()
def export_mermaid(
    *,
    schemapack: Annotated[
        Path,
        typer.Argument(
            help="Provide the path to a schemapack file to export.",
        ),
    ],
    content_properties: Annotated[
        bool,
        typer.Option(
            "--content-properties",
            "-c",
            help="Include properties in the output.",
        ),
    ] = False,
):
    """Generate an entity relationship diagram based on the mermaid markup from the
    provided schemapack.
    """
    with expect_schemapack_errors():
        schemapack_ = load_schemapack(schemapack)

    erd_diagram = export_mermaid_impl(
        schemapack=schemapack_, content_properties=content_properties
    )
    print_output(erd_diagram)


#    print_final_success("Schemapack exported successfully.")



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Test the normalize module."""

from collections.abc import Mapping
from pathlib import Path
from typing import Any

import pytest

from schemapack import denormalize, load_datapack, load_schemapack
from schemapack._internals.spec.custom_types import EmbeddingProfile
from schemapack.exceptions import CircularRelationError
from schemapack.utils import read_json_or_yaml_mapping
from tests.fixtures.examples import (
    DENORMALIZED_CUSTOM_EMBEDDING_PATHS,
    DENORMALIZED_DEEP_EMBEDDING_PATHS,
    VALID_DATAPACK_PATHS,
    VALID_SCHEMAPACK_PATHS,
)


def run_denormalization_test(
    name: str,
    expected_denormalized_path: Path,
    embedding_profile: Mapping[str, Any] | None = None,
):
    """Run the denormalization test with optional ignored relations."""
    schemapack_name = name.split(".", 1)[0]
    schemapack = load_schemapack(VALID_SCHEMAPACK_PATHS[schemapack_name])
    datapack = load_datapack(VALID_DATAPACK_PATHS[name])
    expected_denormalized = read_json_or_yaml_mapping(expected_denormalized_path)
    denormalized = (
        denormalize(
            datapack=datapack,
            schemapack=schemapack,
            embedding_profile=embedding_profile,
        )
        if embedding_profile
        else denormalize(datapack=datapack, schemapack=schemapack)
    )

    assert denormalized == expected_denormalized


@pytest.mark.parametrize(
    "name, expected_denormalized_path",
    DENORMALIZED_DEEP_EMBEDDING_PATHS.items(),
    ids=DENORMALIZED_DEEP_EMBEDDING_PATHS.keys(),
)
def test_denormalize_deep_embedding(name: str, expected_denormalized_path: Path):
    """Test the denormalize function with valid datapacks."""
    run_denormalization_test(name, expected_denormalized_path)


EMBEDDING_PROFILE: Mapping[str, EmbeddingProfile] = {
    "simple_nested_relations": {
        "b": {"c": False}
    },  # direct relation c is not defined but embedded by default
    "rooted_simple_resources": {"files": False},
    "rooted_circular_relations": {"some_relation": {"some_relation": False}},
    "rooted_nested_relations": {
        "some_relation": {}
    },  # this is equivalent of profile {}
    "team_rooted_self_relation": {
        "teammates": {"teammates": False, "manager": False},
        "manager": True,
    },
}


@pytest.mark.parametrize(
    "name, expected_denormalized_path",
    DENORMALIZED_CUSTOM_EMBEDDING_PATHS.items(),
    ids=DENORMALIZED_CUSTOM_EMBEDDING_PATHS.keys(),
)
def test_denormalize_custom_embedding(name: str, expected_denormalized_path: Path):
    """Test the denormalize function with valid datapacks."""
    ignored_relations = EMBEDDING_PROFILE.get(name.rsplit(".", 1)[-1])
    run_denormalization_test(name, expected_denormalized_path, ignored_relations)


@pytest.mark.parametrize(
    "name",
    [
        "self_relation_rooted.rooted_circular_relations",
        "self_relation_rooted.rooted_circular_self_relations",
        "team_rooted.team_rooted_self_relation",
    ],
)
def test_denormalize_circular_relation(name: str):
    """Test the denormalize function fails on datapacks with circular relations."""
    schemapack_name = name.split(".", 1)[0]
    schemapack = load_schemapack(VALID_SCHEMAPACK_PATHS[schemapack_name])
    datapack = load_datapack(VALID_DATAPACK_PATHS[name])

    with pytest.raises(CircularRelationError):
        _ = denormalize(datapack=datapack, schemapack=schemapack)



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""A validation plugin."""

from schemapack._internals.validation.base import ResourceValidationPlugin
from schemapack.exceptions import ValidationPluginError
from schemapack.spec.custom_types import ResourceId
from schemapack.spec.datapack import DataPack, Resource
from schemapack.spec.schemapack import ClassDefinition


class UnknownRelationValidationPlugin(ResourceValidationPlugin):
    """A resource-scoped validation plugin validating that no relation properties
    are present in a resource that don not exist in the schemapack class.
    This only applies to schemapack classes with relations.
    """

    @staticmethod
    def does_apply(*, class_: ClassDefinition) -> bool:
        """A classmethod to check whether this validation plugin is relevant for the
        given schemapack.

        Returns: True if this plugin is relevant for the given class definition.
        """
        return bool(class_.relations)

    def __init__(self, *, class_: ClassDefinition):
        """This plugin is configured with one specific class definition of a schemapack."""
        self._expected_relations = set(class_.relations)

    def validate(
        self, *, resource: Resource, resource_id: ResourceId, datapack: DataPack
    ):
        """Validates a specific resource of the defined class. The entire datapack is
        provided for resolving relations to resources of other classes.

        Raises:
            schemapack.exceptions.ValidationPluginError: If validation fails.
        """
        unknown_relations = {
            relation
            for relation in resource.relations
            if relation not in self._expected_relations
        }

        if unknown_relations:
            raise ValidationPluginError(
                type_="UnknownRelationPropertyError",
                message=(
                    "The following relation propertie(s) is/are unkown: "
                    + ", ".join(unknown_relations)
                ),
                details={
                    "unknown_relations": unknown_relations,
                    "existing_relations": set(resource.relations),
                },
            )



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""A validation plugin."""

from schemapack._internals.validation.base import ResourceValidationPlugin
from schemapack.exceptions import ValidationPluginError
from schemapack.spec.custom_types import ResourceId
from schemapack.spec.datapack import DataPack, Resource
from schemapack.spec.schemapack import ClassDefinition


class MissingMandatoryTargetValidationPlugin(ResourceValidationPlugin):
    """A resource-scoped validation plugin validating that every origin defines at least
    one target resource for relations that are mandatory at the target end.

    This only applies to schemapack classes with relations that are mandatory at the
    target end.
    """

    @staticmethod
    def does_apply(*, class_: ClassDefinition) -> bool:
        """A classmethod to check whether this validation plugin is relevant for the
        given schemapack.

        Returns: True if this plugin is relevant for the given class definition.
        """
        return any(relation.mandatory.target for relation in class_.relations.values())

    def __init__(self, *, class_: ClassDefinition):
        """This plugin is configured with one specific class definition of a schemapack."""
        self._relations_of_interest = {
            relation_name
            for relation_name, relation in class_.relations.items()
            if relation.mandatory.target
        }

    def validate(
        self, *, resource: Resource, resource_id: ResourceId, datapack: DataPack
    ):
        """Validates a specific resource of the defined class. The entire datapack is
        provided for resolving relations to resources of other classes.

        Raises:
            schemapack.exceptions.ValidationPluginError: If validation fails.
        """
        relations_with_missing_targets: set[str] = set()

        for relation_name in self._relations_of_interest:
            try:
                target_resource_ids = resource.relations[relation_name].targetResources
            except KeyError:
                # This is an error but needs to be handled by another validation plugin:
                continue

            if not target_resource_ids:
                relations_with_missing_targets.add(relation_name)

        if relations_with_missing_targets:
            raise ValidationPluginError(
                type_="MissingMandatoryTargetError",
                message=(
                    "No targets were defined for the following mandatory relations: "
                    + ", ".join(relations_with_missing_targets)
                ),
                details={
                    "relations_with_missing_targets": relations_with_missing_targets,
                },
            )



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Utility functions.

Warning: This is an internal part of the library and might change without notice.
"""

import json
import os
from collections.abc import Mapping, Set
from contextlib import contextmanager
from io import StringIO
from pathlib import Path
from typing import Any

import jsonschema
import jsonschema.exceptions
import jsonschema.protocols
import jsonschema.validators
import ruamel.yaml
from arcticfreeze import FrozenDict
from pydantic import BaseModel

from schemapack.exceptions import ParsingError

yaml = ruamel.yaml.YAML(typ="rt")
yaml.indent(mapping=2, sequence=4, offset=2)
yaml.default_flow_style = False


def read_json_or_yaml_mapping(path: Path) -> dict:
    """Reads a JSON object or YAML mapping from file.

    Raises:
        ParsingError:
            If the file cannot be decoded as JSON or YAML or does not contain a
            JSON object or YAML Mapping. Please note, that this parser raise a
            ParsingError for duplicate keys in both JSON objects and YAML mappings.
    """
    with path.open("r", encoding="utf-8") as file:
        try:
            data = yaml.load(file)
        except ruamel.yaml.YAMLError as error:
            raise ParsingError(
                f"The file at '{path}' could not be parsed as JSON or YAML."
            ) from error

    if not isinstance(data, dict):
        raise ParsingError(
            f"The file at '{path}' did not contain a JSON object or a YAML mapping."
        )

    return data


class JsonSchemaError(ValueError):
    """Raised when a JSON schema is invalid."""


def assert_valid_json_schema(schema: Mapping[str, Any]) -> None:
    """Asserts that the given mapping is a valid JSON Schema.

    Raises:
        JsonSchemaError: If the schema is invalid.
    """
    cls: type[jsonschema.protocols.Validator] = jsonschema.validators.validator_for(
        schema
    )
    if isinstance(schema, FrozenDict):
        schema = thaw(schema)
    try:
        cls.check_schema(schema)
    except jsonschema.exceptions.SchemaError as error:
        raise JsonSchemaError(error.message) from error


@contextmanager
def transient_directory_change(path: Path):
    """Change the current working directory temporarily within a with block."""
    original_cwd = os.getcwd()
    os.chdir(path)

    try:
        yield
    finally:
        os.chdir(original_cwd)


def model_to_serializable_dict(
    model: BaseModel,
) -> dict[str, Any]:
    """Converts the provided pydantic model to a JSON-serializable dictionary.

    Returns:
        A dictionary representation of the provided model.
    """
    return json.loads(model.model_dump_json(exclude_defaults=True))


def thaw(frozen: Any) -> Any:
    """Recursively thaws potentially frozen data structures into mutable ones, replacing
    any mappings with dictionaries and tuples with lists.
    """
    if isinstance(frozen, Mapping):
        return {key: thaw(value) for key, value in frozen.items()}
    elif isinstance(frozen, Set):
        return {thaw(item) for item in frozen}
    elif isinstance(frozen, list | tuple):
        return [thaw(item) for item in frozen]
    else:
        return frozen


def dumps_model(
    model: BaseModel,
    *,
    yaml_format: bool = True,
) -> str:
    """Dumps the provided pydantic model as a JSON or YAML-formatted string.

    Args:
        model:
            The model to dump.
        yaml_format:
            Whether to dump as YAML (`True`) or JSON (`False`).
    """
    model_dict = model_to_serializable_dict(model)

    if yaml_format:
        with StringIO() as buffer:
            yaml.dump(model_dict, buffer)
            return buffer.getvalue().strip()

    return json.dumps(model_dict, indent=2)


def write_dict(dict_: dict, *, path: Path, yaml_format: bool) -> None:
    """Writes the provided dictionary to a file at the provided path."""
    if yaml_format:
        with open(path, "w", encoding="utf-8") as file:
            yaml.dump(dict_, file)
    else:
        with open(path, "w", encoding="utf-8") as file:
            json.dump(dict_, file, indent=2)



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Compare two schema definitions for semantic equality."""

from arcticfreeze import FrozenDict
from jsonsubschema import isEquivalent

from schemapack._internals.exceptions import (
    ComparisonError,
    InequivalentContentSchemas,
    InequivalentSchemapacks,
)
from schemapack._internals.spec.custom_types import ContentSchema, RelationPropertyName
from schemapack._internals.spec.schemapack import ClassDefinition, ClassRelation
from schemapack._internals.utils import thaw
from schemapack.spec.schemapack import SchemaPack

__all__ = [
    "assert_equal_schemapack",
    "assert_equivalent_schemapack",
    "is_equal_schemapack",
    "is_equivalent_schemapack",
]


def compare_relations_semantically(
    relations1: FrozenDict[RelationPropertyName, ClassRelation],
    relations2: FrozenDict[RelationPropertyName, ClassRelation],
) -> bool:
    """Compare two relations for semantic equality."""
    if not relations1 and not relations2:
        return True

    if relations1.keys() != relations2.keys():
        return False

    for key in relations1:
        if not compare_class_relations_semantically(relations1[key], relations2[key]):
            return False
    return True


def compare_class_relations_semantically(
    relation1: ClassRelation,
    relation2: ClassRelation,
) -> bool:
    """Compare two ClassRelation objects to determine whether they are equivalent.
    `description` attribute is omitted from the comparison.
    """
    return (
        relation1.targetClass == relation2.targetClass
        and relation1.mandatory == relation2.mandatory
        and relation1.multiple == relation2.multiple
    )


def compare_content_semantically(
    schema1: ContentSchema, schema2: ContentSchema
) -> bool:
    """Compare two content schemas to determine whether they are equivalent."""
    try:
        return isEquivalent(thaw(schema1), thaw(schema2))
    except Exception as exp:
        raise ComparisonError("An error happened while comparing the schemas.") from exp


def compare_class_definitions_semantically(
    class_def1: ClassDefinition,
    class_def2: ClassDefinition,
) -> bool:
    """Compare two ClassDefinition objects to determine whether they are equivalent."""
    return (
        class_def1.id.propertyName == class_def2.id.propertyName
        and compare_relations_semantically(class_def1.relations, class_def2.relations)
        and compare_content_semantically(class_def1.content, class_def2.content)
    )


def assert_equivalent_schemapack(
    schemapack1: SchemaPack, schemapack2: SchemaPack
) -> None:
    """Assert that two schemapacks are equivalent."""
    if schemapack1.rootClass != schemapack2.rootClass:
        raise InequivalentSchemapacks(
            f"Root class mismatch between schemapacks: "
            f"{schemapack1.rootClass} != {schemapack2.rootClass}"
        )

    # Compare class names as sets (order doesn't matter, only membership).
    if schemapack1.classes.keys() != schemapack2.classes.keys():
        difference = schemapack1.classes.keys() ^ schemapack2.classes.keys()
        raise InequivalentSchemapacks(
            f"Class set mismatch between schemapacks for classes: {difference}."
        )

    for class_name in schemapack1.classes:
        if not compare_class_definitions_semantically(
            schemapack1.classes[class_name],
            schemapack2.classes[class_name],
        ):
            raise InequivalentContentSchemas(
                f"Class definition mismatch for '{class_name}'"
            )


def is_equivalent_schemapack(schemapack1: SchemaPack, schemapack2: SchemaPack) -> bool:
    """Check if two schemapacks are equivalent."""
    try:
        assert_equivalent_schemapack(schemapack1, schemapack2)
        return True
    except (InequivalentSchemapacks, InequivalentContentSchemas):
        return False


def assert_equal_schemapack(schemapack1: SchemaPack, schemapack2: SchemaPack) -> None:
    """Assert structural equality of two schemapacks.

    The schemapacks are considered equal if their model instances compare equal
    via '=='. If they are not equal, an 'InequivalentSchemapacks' exception
    is raised.
    """
    if schemapack1 != schemapack2:
        raise InequivalentSchemapacks("Schemapacks are not equal.")


def is_equal_schemapack(schemapack1: SchemaPack, schemapack2: SchemaPack) -> bool:
    """Check structural equality of two schemapacks.

    Returns 'True' if the schemapacks are structurally equal according to
    'assert_equal_schemapack', otherwise returns 'False'.
    """
    try:
        assert_equal_schemapack(schemapack1, schemapack2)
    except InequivalentSchemapacks:
        return False
    return True
